/*
 * Copyright (c) 2013 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-mhn-0.c,v 1.8 2013/10/23 22:13:35 ca Exp $")

#include "sm/debug.h"
#include "sm/error.h"
#include "sm/str.h"
#include "sm/misc.h"
#include "sm/io.h"
#include "sm/test.h"

static int Verbose;

struct ph_S {
	const char *ph_dnsn;
	const char *ph_hn;
	int			ph_res;
};
typedef struct ph_S ph_T;

static ph_T ph_hns[] = {
	{ "*.example.org", "a.b.example.org", -1 },
	{ "www.example.org", "www.example.org", 1 },
	{ "*.example.org", "www.example.org", 1 },
	{ "*.example.org", "server.example.org", 1 },
	{ ".example.org", "www.example.org", -1 },
	{ "www.example.org", ".example.org", 1 },
	{ "w.w.example.org", ".example.org", 1 },
	{ ".example.org", ".example.org", 1 },
	{ "example.org", ".example.org", -1 },
	{ "example.org", ".org", 1 },
	{ "example.org", ".", -1 },
	{ "host.example.org", "my.server.example.org", -1 },
	{ "*.org", "example.org", -1 },
	{ NULL, NULL, 0 }
};

#define BHN_MAX	8

struct mbn_S {
	const char *mbn_dnsn[BHN_MAX];
	const char *mbn_hn;
	int			mbn_res;
};
typedef struct mbn_S mbn_T;

static mbn_T mbn_hns[] = {
	{ { "www.example.org", "*.example.org" }, "server.example.org", 1 },
	{ { "mail.example.org", "www.example.org", "*.example.org" }, "server.example.org", 2 },
	{ { "mail.example.org", "www.example.org", "*.example.org" }, "what.host.example.org", -1 },
	{ { "mail.example.org", "www.example.org", "*.example.org" }, "example.org", -1 },
	{ { "mail.example.org", "www.example.org", "*.example.org" }, "other.org", -1 },
	{ { "mail.example.org", "www.example.org", "*.example.org" }, "some.other.org", -1 },
	{ {NULL}, NULL, 0 }
};


#define CHK2 do { \
	if (ret < 0) ret = -1; else if (ret >= 0) ret = 1; \
	SM_TEST(ph_hns[i].ph_res == ret);	\
	if (Verbose && !(ph_hns[i].ph_res == ret))	\
		sm_io_fprintf(smioerr, "'%S'~'%S'=%d, expected=%d\n", s1, s2, ret, ph_hns[i].ph_res);	\
	} while (0)

#define CHKMNH do { \
	if (ret < 0) ret = -1; \
	SM_TEST(mbn_hns[i].mbn_res == ret);	\
	if (Verbose && !(mbn_hns[i].mbn_res == ret))	\
		sm_io_fprintf(smioerr, "'%S'~'%S'=%d, expected=%d\n", strs[0], s2, ret, mbn_hns[i].mbn_res);	\
	} while (0)

static void
test(void)
{
	int i, r;
	sm_ret_T ret;
	sm_str_P s1, s2, strs[BHN_MAX];

	s1 = sm_str_new(NULL, 64, 256);
	SM_TEST_E(s1 != NULL);
	s2 = sm_str_new(NULL, 64, 256);
	SM_TEST_E(s2 != NULL);
	for (i = 0; i < BHN_MAX; i++) {
		strs[i] = sm_str_new(NULL, 64, 256);
		SM_TEST_E(strs[i] != NULL);
	}

	for (i = 0; ph_hns[i].ph_dnsn != NULL; i++) {
		sm_str_clr(s1);
		sm_str_clr(s2);
		ret = sm_str_scat(s1, ph_hns[i].ph_dnsn);
		SM_TEST_E(sm_is_success(ret));
		ret = sm_str_scat(s2, ph_hns[i].ph_hn);
		SM_TEST_E(sm_is_success(ret));
		r = mta_match_hostname(s1, s2);
		if (r < 0) r = -1; else if (r > 0) r = 1;
		SM_TEST(ph_hns[i].ph_res == r);
		if (Verbose && !(ph_hns[i].ph_res == r))
			sm_io_fprintf(smioerr, "'%S'~'%S'=%d, expected=%d\n", s1, s2, r, ph_hns[i].ph_res);
		strs[0] = s1;
		ret = mta_best_match(strs, 1, s2, false);
		CHK2;
		ret = mta_best_match(strs, 1, s2, true);
		CHK2;
	}

	for (i = 0; mbn_hns[i].mbn_dnsn[0] != NULL; i++) {
		int j;

		for (j = 0; j < BHN_MAX; j++) {
			if (mbn_hns[i].mbn_dnsn[j] == NULL)
				break;
			sm_str_clr(strs[j]);
			ret = sm_str_scat(strs[j], mbn_hns[i].mbn_dnsn[j]);
			SM_TEST_E(sm_is_success(ret));
		}
		sm_str_clr(s2);
		ret = sm_str_scat(s2, mbn_hns[i].mbn_hn);
		SM_TEST_E(sm_is_success(ret));
		ret = mta_best_match(strs, j, s2, false);
		CHKMNH;
		ret = mta_best_match(strs, j, s2, true);
		CHKMNH;
	}


  error:
	return;
}

int
main(int argc, char *argv[])
{
	int c;

	opterr = 0;
	Verbose = 0;
	while ((c = getopt(argc, argv, "V")) != -1) {
		switch (c) {
		  case 'V':
			++Verbose;
			break;
		  default:
			exit(1);
		}
	}
	sm_test_begin(argc, argv, "test cert hostname matching");
	argc -= optind;
	argv += optind;
	test();
	return sm_test_end();
}
