/*
 * Copyright (c) 2004, 2005, 2010 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-wait-0.c,v 1.23 2013/11/21 14:05:34 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/sysexits.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/misc.h"
#include "sm/test.h"
#include "sm/evthr.h"
#include "sm/queue.h"

#include <stdio.h>

/* for debugging */
sm_evthr_task_P TASK;

#define WAIT_TASKS_MAX	32

#define T_THR_MIN	1
#define T_S_THR_MAX	2
#define T_SRV_DELAY	20
#define QUERY_TMO	2

/*

2004-10-23	priority: 8
libevthr: test:  minthr=1, maxthr=2
start one task twice which
- starts another task that will eventually signal a cv
- returns "async" and waits for a cv
(as "sleep" task, 1s each?)
the "other" task should act like DNS lookups:
- create a request
- after a while answer the request
i.e., there is one task which reads requests from a queue and
invoke some callback function with some context;
creating a request means appending a structure to the list,
the struct contains the callback function and the context (some id?
in this case the cv should be sufficient)

2004-10-24
system can hang if "can block" is set for a task
because libevthr skips those when creating more threads
and it doesn't "remember" that another task (non-blocking)
wasn't created before?

*/

static int Verbose = 0;
static uint Dbglvl = 0;
static sm_evthr_ctx_P evthr_ctx;
static uint s_delay = QUERY_TMO;

/* "server" context */
struct s_ctx_S
{
	pthread_mutex_t			 s_mutex;
	timeval_T				 s_nextt;
	TAILQ_HEAD(, r_req_S)	 s_head;	/* list of requests */
};
typedef struct s_ctx_S	s_ctx_T, *s_ctx_P;

/* "wait" context */
struct w_ctx_S
{
	s_ctx_P			 w_s_ctx;	/* "server" context */
	sm_evthr_task_P	 w_s_tsk;	/* "server" task */
	uint			 w_idx;
	int				 w_reqs;
	int				 w_count;
	int				 w_nresults;
	int				 w_sleep;
	timeval_T		 w_nextt;
};
typedef struct w_ctx_S	w_ctx_T, *w_ctx_P;

/* request/result context */
struct a_res_S
{
	pthread_mutex_t	 a_mutex;
	pthread_cond_t	 a_cond;
	int				 a_result;
	long			 a_rqt;
};
typedef struct a_res_S	a_res_T, *a_res_P;

/* request list */
struct r_req_S
{
	a_res_P					 r_a_res;	/* pointer to result context */
	TAILQ_ENTRY(r_req_S)	 r_link;
};
typedef struct r_req_S	r_req_T, *r_req_P;

int Pid_fd = -1;
static char Pid_file[] = "t-wait-0.pid";

/*
**  MYEXIT -- clean up and exit
**
**	Parameters:
**		rc -- return code
**
**	Returns:
**		doesn't.
*/

static void
myexit(int rc)
{
	static int fd = 0;

	if (Verbose > 2)
		fprintf(stderr, "func=myexit, Pid_fd=%d, rc=%d, fd=%d\n", Pid_fd, rc, fd);

	/* try to minimize race condition: multiple callers! use mutex? */
	if (fd < 0 || (fd = Pid_fd) < 0) return;
	fd = -1;
	close(Pid_fd);
	Pid_fd = -1;
	(void) unlink(Pid_file);
	exit(rc);
}

/*
**  W_NEW -- create new "wait" context
**
**	Parameters:
**		pw_ctx -- (pointer to) "wait" context (output)
**		idx -- index
**		reqs -- number of requests
**		s_tsk -- "server" task
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
w_new(w_ctx_P *pw_ctx, uint idx, int reqs, sm_evthr_task_P s_tsk)
{
	w_ctx_P w_ctx;

	SM_REQUIRE(pw_ctx != NULL);
	w_ctx = (w_ctx_P) sm_zalloc(sizeof(*w_ctx));
	if (w_ctx == NULL)
		return sm_error_temp(SM_EM_EVTHR, ENOMEM);
	w_ctx->w_s_tsk = s_tsk;
	w_ctx->w_idx = idx;
	w_ctx->w_reqs = reqs;
	w_ctx->w_sleep = 0;
	*pw_ctx = w_ctx;
	return SM_SUCCESS;
}

/*
**  A_NEW -- create new "result" context
**
**	Parameters:
**		pa_ctx -- (pointer to) "result" context (output)
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
a_new(a_res_P *pa_res)
{
	int r;
	a_res_P a_res;

	SM_REQUIRE(pa_res != NULL);
	a_res = (a_res_P) sm_zalloc(sizeof(*a_res));
	if (a_res == NULL)
		return sm_error_temp(SM_EM_EVTHR, ENOMEM);
	r = pthread_mutex_init(&a_res->a_mutex, NULL);
	SM_TEST(r == 0);
	if (r != 0)
		goto error;
	r = pthread_cond_init(&a_res->a_cond, NULL);
	SM_TEST(r == 0);
	if (r != 0)
		goto error;
	*pa_res = a_res;
	return SM_SUCCESS;
  error:
	return sm_error_perm(SM_EM_EVTHR, r);
}

/*
**  S_NEW -- create new "server" context
**
**	Parameters:
**		ps_ctx -- (pointer to) "server" context (output)
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
s_new(s_ctx_P *ps_ctx)
{
	int r;
	s_ctx_P s_ctx;

	SM_REQUIRE(ps_ctx != NULL);
	s_ctx = (s_ctx_P) sm_zalloc(sizeof(*s_ctx));
	if (s_ctx == NULL)
		return sm_error_temp(SM_EM_EVTHR, ENOMEM);
	r = pthread_mutex_init(&s_ctx->s_mutex, NULL);
	SM_TEST(r == 0);
	if (r != 0)
		goto error;
	TAILQ_INIT(&s_ctx->s_head);
	*ps_ctx = s_ctx;
	return SM_SUCCESS;
  error:
	return sm_error_perm(SM_EM_EVTHR, r);
}

/*
**  R_NEW -- create new "request" list
**
**	Parameters:
**		pw_ctx -- (pointer to) "request" list (output)
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
r_new(r_req_P *pr_req)
{
	r_req_P r_req;

	SM_REQUIRE(pr_req != NULL);
	r_req = (r_req_P) sm_zalloc(sizeof(*r_req));
	if (r_req == NULL)
		return sm_error_temp(SM_EM_EVTHR, ENOMEM);
	*pr_req = r_req;
	return SM_SUCCESS;
}

/*
**  NOTIFY -- signal condition so waiting thread can get result
**
**	Parameters:
**		a_res -- result context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
notify(a_res_P a_res)
{
	sm_ret_T ret;
	int r;

	SM_REQUIRE(a_res != NULL);
	ret = SM_SUCCESS;

	if (Verbose > 2)
		fprintf(stderr, "func=notify, a_res=%p\n", a_res);
	r = pthread_mutex_lock(&a_res->a_mutex);
	if (Verbose > 5)
		fprintf(stderr, "func=notify, a_res=%p, lock=%d\n", a_res, r);
	SM_TEST(r == 0);
	if (r != 0)
		return sm_error_perm(SM_EM_EVTHR, r);
	r = pthread_cond_signal(&a_res->a_cond);
	if (Verbose > 0)
		fprintf(stderr, "func=notify, a_res=%p, cond_signal=%d\n", a_res, r);
	SM_TEST(r == 0);
	r = pthread_mutex_unlock(&a_res->a_mutex);
	SM_TEST(r == 0);
	return ret;
}

/*
**  GET_RESULT -- get (wait for) result
**
**	Parameters:
**		a_res -- result context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
get_result(a_res_P a_res, uint idx)
{
	int r;

	if (Verbose > 3)
		fprintf(stderr, "func=get_result, idx=%u, a_res=%p\n", idx, a_res);
	r = pthread_mutex_lock(&a_res->a_mutex);
	if (r != 0) {
		/* LOG something! */
		return sm_error_perm(SM_EM_EVTHR, r);
	}

	if (a_res->a_result == 0) {
		while (a_res->a_result == 0) {
			if (Verbose > 1)
				fprintf(stderr,
					"func=get_result, idx=%u, a_res=%p, where=before_cond_wait\n"
					, idx, a_res);
			(void) evthr_before_block(evthr_ctx);
			r = pthread_cond_wait(&a_res->a_cond, &a_res->a_mutex);
			(void) evthr_after_block(evthr_ctx);
			if (Verbose > 1)
				fprintf(stderr, "func=get_result, idx=%u, a_res=%p, wait=%d\n"
					, idx, a_res, r);
			if (r == EINVAL)
				break;
			if (r == 0) {
				r = pthread_mutex_unlock(&a_res->a_mutex);
				SM_ASSERT(r == 0);
			}
		}
	}
	else {
		r = pthread_mutex_unlock(&a_res->a_mutex);
		SM_ASSERT(r == 0);
	}
	if (r == 0)
		return SM_SUCCESS;
	return sm_error_perm(SM_EM_EVTHR, r);
}

/*
**  ADD_REQUEST -- add a request
**
**	Parameters:
**		s_ctx -- server context
**		a_res -- result context
**		now -- current time
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
add_request(s_ctx_P s_ctx, sm_evthr_task_P s_tsk, a_res_P a_res, timeval_T now)
{
	sm_ret_T ret;
	int r;
	r_req_P r_req;
	timeval_T sleept, delay;

	SM_REQUIRE(s_ctx != NULL);
	SM_REQUIRE(a_res != NULL);
	ret = r_new(&r_req);
	SM_TEST(ret == SM_SUCCESS);
	if (sm_is_err(ret))
		return ret;
	r_req->r_a_res = a_res;
	a_res->a_rqt = now.tv_sec + s_delay;

	delay.tv_sec = s_delay;
	delay.tv_usec = 0;
	timeradd(&now, &delay, &sleept);
	ret = evthr_new_sl(s_tsk, sleept, false);

	if (Verbose > 2)
		fprintf(stderr, "func=add_request, s_ctx=%p, a_res=%p\n", s_ctx, a_res);
	r = pthread_mutex_lock(&s_ctx->s_mutex);
	SM_TEST(r == 0);
	if (r != 0)
		return sm_error_perm(SM_EM_EVTHR, r);

	TAILQ_INSERT_TAIL(&s_ctx->s_head, r_req, r_link);

	r = pthread_mutex_unlock(&s_ctx->s_mutex);
	SM_TEST(r == 0);
	return ret;
}

/*
**  FCTW -- "worker"/"wait" function
**
**	Parameters:
**		tsk -- evthr task
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
fctw(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	w_ctx_P w_ctx;
	timeval_T sleept, delay;

	SM_ASSERT(tsk != NULL);
	w_ctx = (w_ctx_P) tsk->evthr_t_actx;
	SM_ASSERT(w_ctx != NULL);
	++w_ctx->w_count;
	ret = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(ret == 0);
	if (Verbose > 2)
		fprintf(stderr, "func=fctw, now=%ld, w_idx=%u, count=%d\n"
			, sleept.tv_sec, w_ctx->w_idx, w_ctx->w_count);
	if (w_ctx->w_count < w_ctx->w_reqs) {
		a_res_P a_res;

		ret = a_new(&a_res);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			return ret;
		ret = add_request(w_ctx->w_s_ctx, w_ctx->w_s_tsk, a_res, sleept);
		SM_TEST(sm_is_success(ret));
		if (Verbose > 4)
			fprintf(stderr, "func=fctw, now=%ld, w_idx=%u"
				", a_res=%p"
				", evthr_t_rqevf=%x"
				", evthr_t_evocc=%x"
				", evthr_t_state=%x"
				", evthr_t_flags=%x"
				", evthr_t_sleep=%ld"
				"\n"
				, sleept.tv_sec
				, w_ctx->w_idx
				, a_res
				, tsk->evthr_t_rqevf
				, tsk->evthr_t_evocc
				, tsk->evthr_t_state
				, tsk->evthr_t_flags
				, tsk->evthr_t_sleep.tv_sec
				);

		ret = evthr_waitq_app(tsk);
		SM_TEST(sm_is_success(ret));
		if (Verbose > 3)
			fprintf(stderr,
				"func=fctw, now=%ld, w_idx=%u, evthr_waitq_app=%x\n"
				, sleept.tv_sec, w_ctx->w_idx, ret);
		ret = evthr_new_sl(tsk, sleept, true);
		if (Verbose > 3)
			fprintf(stderr,
				"func=fctw, now=%ld, w_idx=%u, evthr_new_sl=%x\n"
				, sleept.tv_sec, w_ctx->w_idx, ret);
		if (w_ctx->w_sleep > 0)
			sleep(w_ctx->w_sleep);
		ret = get_result(a_res, w_ctx->w_idx);
		SM_TEST(sm_is_success(ret));
		++w_ctx->w_nresults;
		ret = evthr_timeval(evthr_ctx, &sleept);
		SM_TEST(ret == 0);
		if (Verbose > 0)
			fprintf(stderr,
				"func=fctw, now=%ld, w_idx=%u, get_result=%x, a_result=%d\n"
				, sleept.tv_sec, w_ctx->w_idx, ret, a_res->a_result);
		if (w_ctx->w_nresults + 1 >= w_ctx->w_reqs) {
			if (Verbose > 0)
				fprintf(stderr,
					"func=fctw, now=%ld, w_idx=%u, n_results=%d, stat=exiting\n"
					, sleept.tv_sec, w_ctx->w_idx, w_ctx->w_nresults);
			myexit(0);	/* HACK... */
			/* NOTREACHED */
		}
		return EVTHR_OK;
	}

	delay.tv_sec = 20;
	delay.tv_usec = 0;
	timeradd(&sleept, &delay, &tsk->evthr_t_sleep);
	w_ctx->w_nextt = tsk->evthr_t_sleep;
	if (Verbose > 2)
		fprintf(stderr,
			"func=fctw, now=%ld, w_idx=%u, next=%lds\n"
			, sleept.tv_sec, w_ctx->w_idx, delay.tv_sec);
	return EVTHR_SLPQ;
}

/*
**  FCTBL -- "wait" (blocking) function
**
**	Parameters:
**		tsk -- evthr task
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
fctbl(sm_evthr_task_P tsk)
{
	sm_ret_T ret;
	w_ctx_P w_ctx;
	timeval_T now;

	SM_ASSERT(tsk != NULL);
	w_ctx = (w_ctx_P) tsk->evthr_t_actx;
	SM_ASSERT(w_ctx != NULL);
	ret = evthr_timeval(evthr_ctx, &now);
	SM_TEST(ret == 0);
	if (Verbose > 3)
		fprintf(stderr, "func=fctbl, now=%ld, w_idx=%u, count=%d\n"
			, now.tv_sec, w_ctx->w_idx, w_ctx->w_count);
	while (w_ctx->w_count++ < w_ctx->w_reqs) {
		a_res_P a_res;

		ret = a_new(&a_res);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			return ret;
		ret = add_request(w_ctx->w_s_ctx, w_ctx->w_s_tsk, a_res, now);
		SM_TEST(sm_is_success(ret));
		if (Verbose > 4)
			fprintf(stderr, "func=fctbl now=%ld, w_idx=%u"
				", w_count=%d"
				", w_reqs=%d"
				", a_res=%p"
				", evthr_t_rqevf=%X"
				", evthr_t_evocc=%X"
				", evthr_t_state=%X"
				", evthr_t_flags=%X"
				", evthr_t_sleep=%ld"
				"\n"
				, now.tv_sec
				, w_ctx->w_idx
				, w_ctx->w_count
				, w_ctx->w_reqs
				, a_res
				, tsk->evthr_t_rqevf
				, tsk->evthr_t_evocc
				, tsk->evthr_t_state
				, tsk->evthr_t_flags
				, tsk->evthr_t_sleep.tv_sec
				);

		ret = get_result(a_res, w_ctx->w_idx);
		SM_TEST(sm_is_success(ret));
		++w_ctx->w_nresults;
		ret = evthr_timeval(evthr_ctx, &now);
		SM_TEST(ret == 0);
		if (Verbose > 0)
			fprintf(stderr,
				"func=fctbl, now=%ld, w_idx=%u, get_result=%X, a_result=%d\n"
				, now.tv_sec, w_ctx->w_idx, ret, a_res->a_result);
	}
	if (Verbose > 0)
		fprintf(stderr,
			"func=fctbl, now=%ld, w_idx=%u, stat=terminating\n"
			, now.tv_sec, w_ctx->w_idx);
	myexit(0); /* return EVTHR_TERM; */
	/* NOTREACHED */
	return EVTHR_TERM;
}

/*
**  FCTSRV -- "server" function
**
**	Parameters:
**		tsk -- evthr task
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
fctsrv(sm_evthr_task_P tsk)
{
	int r;
	uint cnt;
	bool empty;
	s_ctx_P s_ctx;
	timeval_T sleept, delay;
	r_req_P r_req;
	static int result = 0;
	TAILQ_HEAD(, r_req_S)	 n_head;

	SM_ASSERT(tsk != NULL);
	SM_TEST(evthr_got_slp(tsk));
	s_ctx = (s_ctx_P) tsk->evthr_t_actx;
	SM_ASSERT(s_ctx != NULL);
	r = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(r == 0);
	delay.tv_sec = 2;
	delay.tv_usec = 0;
	cnt = 0;
	empty = true;

	TAILQ_INIT(&n_head);
	if (Verbose > 2)
		fprintf(stderr, "func=fctsrv, now=%ld, s_ctx=%p\n"
			, sleept.tv_sec, s_ctx);
	r = pthread_mutex_lock(&s_ctx->s_mutex);
	SM_TEST_E(r == 0);
	if (r != 0)
		goto error;

	/* move all "timed out" requests into a local list */
	while (!TAILQ_EMPTY(&s_ctx->s_head)) {
		r_req = TAILQ_FIRST(&s_ctx->s_head);

		if (sleept.tv_sec < r_req->r_a_res->a_rqt) {
			delay.tv_sec = r_req->r_a_res->a_rqt - sleept.tv_sec;
			if (delay.tv_sec == 0)
				delay.tv_sec = 10;
			break;
		}

		TAILQ_REMOVE(&s_ctx->s_head, r_req, r_link);
		if (Verbose > 1)
			fprintf(stderr, "func=fctsrv, now=%ld, s_ctx=%p, r_req=%p\n"
				, sleept.tv_sec, s_ctx, r_req);
		TAILQ_INSERT_TAIL(&n_head, r_req, r_link);
		r_req->r_a_res->a_result = ++result;
	}
	r = pthread_mutex_unlock(&s_ctx->s_mutex);
	SM_TEST(r == 0);

	/* use local list for notifications */
	while (!TAILQ_EMPTY(&n_head)) {
		r_req = TAILQ_FIRST(&n_head);
		TAILQ_REMOVE(&n_head, r_req, r_link);
		notify(r_req->r_a_res);
		++cnt;
	}
	if (Verbose > 3)
		fprintf(stderr, "func=fctsrv, now=%ld, cnt=%u\n", sleept.tv_sec, cnt);

  error:
	r = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(r == 0);
	timeradd(&sleept, &delay, &tsk->evthr_t_sleep);
	s_ctx->s_nextt = tsk->evthr_t_sleep;
	return EVTHR_SLPQ;
}

/*
**  FCTSTOP -- stop function: exits.
**
**	Parameters:
**		tsk -- evthr task
**
**	Returns:
**		never.
*/

static sm_ret_T
fctstop(sm_evthr_task_P tsk)
{
	int r;
	timeval_T sleept;

	SM_ASSERT(tsk != NULL);
	SM_ASSERT(evthr_got_slp(tsk));
	r = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(r == 0);
	fprintf(stderr, "func=fctstop, now=%ld, state=ERROR\n", sleept.tv_sec);
	myexit(1);
	/* NOTREACHED */
	return(1);
}

/*
**  TESTWAIT -- set up test system
**
**	Parameters:
**		canblock -- mark "wait" thread as "can block"?
**		blocking -- run ""block" function instead of "wait" function?
**		minthr -- min threads (evthr)
**		maxthr -- max threads (evthr)
**		hard -- hard limit for number of threads
**		maxwtsks -- max "wait" threads
**
**	Returns:
**		usual sm_error code
*/

static void
testwait(bool canblock, bool blocking, uint reqs, uint minthr, uint maxthr, uint hard, uint maxwtsks)
{
	sm_ret_T ret;
	uint u;
	sm_evthr_task_P	srvtsk, stoptask;
	sm_evthr_task_P	wtask[WAIT_TASKS_MAX];
	w_ctx_P w_ctx[WAIT_TASKS_MAX];
	s_ctx_P s_ctx1;
	timeval_T sleept, nextt, delay;

	evthr_ctx = NULL;
	ret = thr_init();
	SM_TEST(sm_is_success(ret));
	sm_memzero(&sleept, sizeof(sleept));
	ret = evthr_init(&evthr_ctx, minthr, maxthr, 2);
	SM_TEST(sm_is_success(ret));
	SM_TEST(evthr_ctx != NULL);
	if (sm_is_err(ret))
		goto err2;
	evthr_set_dbglvl(evthr_ctx, Dbglvl);
	if (hard > 0) {
		ret = evthr_set_max_h(evthr_ctx, hard);
		SM_TEST(sm_is_success(ret));
	}

	/* "request server" function: sleep 20s before doing something */
	s_ctx1 = NULL;	/* shut up bogus gcc compiler warning */
	ret = s_new(&s_ctx1);
	SM_TEST(sm_is_success(ret));
	SM_TEST(s_ctx1 != NULL);
	if (sm_is_err(ret))
		goto error;
	ret = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(ret == 0);
	sleept.tv_sec += T_SRV_DELAY;
	s_ctx1->s_nextt = sleept;
	ret = evthr_task_new(evthr_ctx, &srvtsk, EVTHR_EV_SL, -1, &sleept,
				fctsrv, (void *) s_ctx1);
	SM_TEST(sm_is_success(ret));
	SM_TEST(srvtsk != NULL);
	if (sm_is_err(ret))
		goto error;
	TASK = srvtsk;
	if (Verbose > 3)
		fprintf(stderr, "func=testwait, srv_task=%p\n", srvtsk);

	/* "block" function: sleep a bit before doing something */
	for (u = 0; u < maxwtsks; u++) {
		ret = w_new(&w_ctx[u], u, reqs, srvtsk);
		SM_TEST(sm_is_success(ret));
		SM_TEST(w_ctx[u] != NULL);
		if (sm_is_err(ret))
			goto error;
		w_ctx[u]->w_s_ctx = s_ctx1;
		ret = evthr_timeval(evthr_ctx, &sleept);
		SM_TEST(ret == 0);
		delay.tv_sec = 0;
		delay.tv_usec = 1000;
		timeradd(&sleept, &delay, &nextt);
		w_ctx[u]->w_nextt = nextt;
		ret = evthr_task_new(evthr_ctx, &wtask[u], EVTHR_EV_SL, -1, &nextt,
					blocking ? fctbl : fctw, (void *) w_ctx[u]);
		SM_TEST(sm_is_success(ret));
		SM_TEST(wtask[u] != NULL);
		if (sm_is_err(ret))
			goto error;
		if (canblock) {
			EVTHRT_SET_FLAG(wtask[u], EVTHRT_FL_BLOCK);
			if (Verbose > 3)
				fprintf(stderr, "func=testwait, idx=%u, wait_task=%p, flags=%x\n"
					, u, wtask[u], wtask[u]->evthr_t_flags);
		}
	}

	/*
	 *  "stop" function: sleep 60s before stopping system (emergency stop).
	 *  Note: this can only stop the system if the task can actually run,
	 *  which is not the case if a deadlock (no available threads) occurs.
	 */

	ret = evthr_timeval(evthr_ctx, &sleept);
	SM_TEST(ret == 0);
	delay.tv_sec = 60;
	delay.tv_usec = 0;
	timeradd(&sleept, &delay, &nextt);
	ret = evthr_task_new(evthr_ctx, &stoptask, EVTHR_EV_SL, -1, &nextt,
				fctstop, (void *) NULL);
	SM_TEST(sm_is_success(ret));
	SM_TEST(stoptask != NULL);
	if (sm_is_err(ret))
		goto error;

	ret = evthr_loop(evthr_ctx);
	SM_TEST(sm_is_success(ret));
	if (!sm_is_success(ret))
		fprintf(stderr, "evthr_loop=%x\n", ret);

  error:
	if (evthr_ctx != NULL) {
		ret = evthr_stop(evthr_ctx, 0);
		SM_TEST(sm_is_success(ret));
		if (!sm_is_success(ret))
			fprintf(stderr, "evthr_stop=%x\n", ret);
		evthr_ctx = NULL;
	}
  err2:
	ret = thr_stop();
	SM_TEST(sm_is_success(ret));
}

/*
**  USAGE -- show usage
**
**	Parameters:
**		prt -- name of program
**
**	Returns:
**		nothing
*/

static void
usage(const char *prg)
{
	fprintf(stderr, "usage: %s [options]\n", prg);
	fprintf(stderr, "options:\n");
	fprintf(stderr, "-B      use \"block\" function\n");
	fprintf(stderr, "-b      set canblock for \"wait\" threads\n");
	fprintf(stderr, "-d n    debug level\n");
	fprintf(stderr, "-h n    hard limit\n");
	fprintf(stderr, "-M n    max threads [%d]\n", T_THR_MIN);
	fprintf(stderr, "-m n    min threads [%d]\n", T_S_THR_MAX);
	fprintf(stderr, "-s n    delay for server ( [%ds])\n", QUERY_TMO);
	fprintf(stderr, "-V      increase verbosity\n");
	fprintf(stderr, "-w n    create n \"wait\" tasks\n");
	myexit(EX_USAGE);
}

/*
**  MAIN -- main
**
**	Parameters:
**		argc -- arg counter
**		argv -- arg vector
**
**	Returns:
**		test result
*/

int
main(int argc, char *argv[])
{
	int c;
	uint hard, minthr, maxthr, maxwtsks, reqs;
	bool canblock, blocking;
	char errtxt[256];

	canblock = blocking = false;
	hard = 0;
	reqs = 3;
	minthr = T_THR_MIN;
	maxthr = T_S_THR_MAX;
	maxwtsks = 1;
	while ((c = getopt(argc, argv, "Bbd:h:M:m:r:s:w:V")) != -1) {
		switch (c) {
		  case 'B':
			blocking = true;
			break;
		  case 'b':
			canblock = true;
			break;
		  case 'd':
			Dbglvl = atoi(optarg);
			break;
		  case 'h':
			hard = (uint) atoi(optarg);
			break;
		  case 'M':
			maxthr= (uint) atoi(optarg);
			break;
		  case 'm':
			minthr= (uint) atoi(optarg);
			break;
		  case 'r':
			reqs = (uint) atoi(optarg);
			break;
		  case 's':
			s_delay = (uint) atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  case 'w':
			maxwtsks = (uint) atoi(optarg);
			break;
		  default:
			usage(argv[0]);
			return(EX_USAGE);
		}
	}
	if (hard > 0 && maxthr > 0 && hard <= maxthr)
		usage(argv[0]);
	if (maxwtsks <= 0 || maxwtsks >= WAIT_TASKS_MAX)
		usage(argv[0]);
	if (s_delay <= 0)
		usage(argv[0]);

	sm_test_begin(argc, argv, "test wait with deadlock avoidance");
	c = sm_chk_pidfile(Pid_file, &Pid_fd, errtxt, sizeof(errtxt));
	SM_TEST(0 == c);
	if (0 == c)
		testwait(canblock, blocking, reqs, minthr, maxthr, hard, maxwtsks);
	else
		fprintf(stderr, "error: %s\n", errtxt);
	return sm_test_end();
}
