/* based on: */

/*	$OpenBSD: inet_pton.c,v 1.7 2006/12/30 23:37:37 itojun Exp $	*/

/* Copyright (c) 1996 by Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
 * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */


#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm_inet_a2ipv6.c,v 1.4 2013/10/01 03:52:50 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/net.h"
#include "sm/nameser.h"	/* some INADDR constants */

#if MTA_NETINET6
/* int
 * inet_pton4(src, dst)
 *	like inet_aton() but without all the hexadecimal and shorthand.
 * return:
 *	1 if `src' is a valid dotted quad, else 0.
 * notice:
 *	does not touch `dst' unless it's returning 1.
 * author:
 *	Paul Vixie, 1996.
 */
static int
inet_pton4(const char *src, uchar *dst)
{
	static const char digits[] = "0123456789";
	int saw_digit, octets, ch;
	uchar tmp[INADDRSZ], *tp;

	saw_digit = 0;
	octets = 0;
	*(tp = tmp) = 0;
	while ((ch = *src++) != '\0') {
		const char *pch;

		if ((pch = strchr(digits, ch)) != NULL) {
			uint new = *tp * 10 + (pch - digits);

			if (new > 255)
				return (0);
			if (! saw_digit) {
				if (++octets > 4)
					return (0);
				saw_digit = 1;
			}
			*tp = new;
		} else if (ch == '.' && saw_digit) {
			if (octets == 4)
				return (0);
			*++tp = 0;
			saw_digit = 0;
		} else
			return (0);
	}
	if (octets < 4)
		return (0);

	memcpy(dst, tmp, INADDRSZ);
	return (1);
}

/*
**  SM_INET_A2IPV6 -- (check and) convert ASCII version of IPv6 address to binary
**
**	Parameters:
**		src -- begin of ASCII representation
**		end -- (pointer to) end of ASCII representation (output)
**		ipv6 -- IPv6 address (output)
**
**	Returns:
**		usual sm_error code; ERANGE, EINVAL, ...
*/

/* int
 * inet_pton6(src, dst)
 *	convert presentation level address to network order binary form.
 * return:
 *	1 if `src' is a valid [RFC1884 2.2] address, else 0.
 * notice:
 *	(1) does not touch `dst' unless it's returning 1.
 *	(2) :: in a full address is silently ignored.
 * credit:
 *	inspired by Mark Andrews.
 * author:
 *	Paul Vixie, 1996.
 */

sm_ret_T
sm_inet_a2ipv6(const char *src, char **end, ipv6_T *ipv6)
{
	static const char xdigits_l[] = "0123456789abcdef",
			  xdigits_u[] = "0123456789ABCDEF";
	uchar tmp[IN6ADDRSZ], *tp, *endp, *colonp;
	const char *xdigits, *curtok;
	int ch, saw_xdigit, count_xdigit;
	uint val;
	bool gotend, brackets;

	if (end != NULL)
		*end = NULL;

	memset(tmp, '\0', sizeof(tmp));
	tp = tmp;
	endp = tp + IN6ADDRSZ;
	colonp = NULL;
	brackets = (*src == '[');
	if (brackets)
		++src;
	/* Leading :: requires some special handling. */
	if (*src == ':')
		if (*++src != ':')
			return sm_error_perm(SM_EM_IP, EINVAL);

	gotend = false;

	curtok = src;
	saw_xdigit = count_xdigit = 0;
	val = 0;
	while ((ch = *src++) != '\0') {
		const char *pch;

		if ((pch = strchr((xdigits = xdigits_l), ch)) == NULL)
			pch = strchr((xdigits = xdigits_u), ch);
		if (pch != NULL) {
			if (count_xdigit >= 4)
				return sm_error_perm(SM_EM_IP, EINVAL);
			val <<= 4;
			val |= (pch - xdigits);
			if (val > 0xffff)
				return sm_error_perm(SM_EM_IP, ERANGE);
			saw_xdigit = 1;
			count_xdigit++;
			continue;
		}
		if (ch == ':') {
			curtok = src;
			if (!saw_xdigit) {
				if (colonp)
					return sm_error_perm(SM_EM_IP, EINVAL);
				colonp = tp;
				continue;
			}
			else if (*src == '\0')
				return sm_error_perm(SM_EM_IP, EINVAL);
			if (tp + INT16SZ > endp)
				return sm_error_perm(SM_EM_IP, EINVAL);
			*tp++ = (uchar) (val >> 8) & 0xff;
			*tp++ = (uchar) val & 0xff;
			saw_xdigit = 0;
			count_xdigit = 0;
			val = 0;
			continue;
		}
		if (ch == '.' && ((tp + INADDRSZ) <= endp) && inet_pton4(curtok, tp) > 0) {
			tp += INADDRSZ;
			saw_xdigit = 0;
			count_xdigit = 0;
			break;	/* '\0' was seen by inet_pton4(). */
		}
		if (brackets && ch == ']')
			break;

		/* any other char: end? */
		break;
	}
	if (brackets) {
		if (ch != ']')
			return sm_error_perm(SM_EM_IP, EINVAL);
		else
			++src;
	}
	if (saw_xdigit) {
		if (tp + INT16SZ > endp)
			return sm_error_perm(SM_EM_IP, EINVAL);
		*tp++ = (uchar) (val >> 8) & 0xff;
		*tp++ = (uchar) val & 0xff;
	}
	if (colonp != NULL) {
		/*
		 * Since some memmove()'s erroneously fail to handle
		 * overlapping regions, we'll do the shift by hand.
		 */
		const int n = tp - colonp;
		int i;

		for (i = 1; i <= n; i++) {
			endp[- i] = colonp[n - i];
			colonp[n - i] = 0;
		}
		tp = endp;
	}
	if (tp != endp)
		return sm_error_perm(SM_EM_IP, EINVAL);
	memcpy(ipv6, tmp, IN6ADDRSZ);
	if (end != NULL)
		*end = (char *)src - 1;
	return SM_SUCCESS;
}
#endif /* MTA_NETINET6 */
